"""Provides a simple func for getting resources."""

from importlib.resources import as_file, files
from pathlib import Path
from types import ModuleType

from pyrig.dev.artifacts import resources
from pyrig.src.modules.module import get_same_modules_from_deps_depen_on_dep


def get_resource_path(name: str, package: ModuleType) -> Path:
    """Get the path to a resource.

    Uses importlib.resources to get the path to a resource.
    Also usefule bc it handles MEIPASS for us.

    Args:
        name: The name of the resource with extension
        package: The package that contains the resource

    Returns:
        The path to the resource
    """
    resource_path = files(package) / name
    with as_file(resource_path) as path:
        return path


def get_all_resources_pkgs_from_deps_depen_on_dep(
    dep: ModuleType,
) -> list[ModuleType]:
    """Get all resources packages from dependencies depending on dep.

    Args:
        dep: The dependency to find resources packages from

    Returns:
        A list of resources packages
    """
    return get_same_modules_from_deps_depen_on_dep(resources, dep)



"""Config utilities for experiment.py."""

from pathlib import Path

from pyrig.dev.configs.base.base import PythonConfigFile


class ExperimentConfigFile(PythonConfigFile):
    """Config file for experiment.py.

    Is at root level and in .gitignore for experimentation.
    """

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_content_str(cls) -> str:
        """Get the config."""
        return '''"""This file is for experimentation and is ignored by git."""
'''



"""Testing utilities for introspection and validation.

This module provides utility functions for working with tests, including:
- Asserting that all objects in the source have corresponding test objects
- Generating the content for a conftest.py file

Returns:
    Various utility functions for testing introspection and validation.

"""

import logging
from collections.abc import Callable
from types import ModuleType
from typing import Any

from pyrig.src.modules.module import (
    get_objs_from_obj,
    make_obj_importpath,
)
from pyrig.src.testing.assertions import assert_with_msg
from pyrig.src.testing.convention import (
    get_obj_from_test_obj,
    make_test_obj_importpath_from_obj,
    make_untested_summary_error_msg,
)
from pyrig.src.testing.create_tests import create_tests

logger = logging.getLogger(__name__)


def assert_no_untested_objs(
    test_obj: ModuleType | type | Callable[..., Any],
) -> None:
    """Assert that all objects in the source have corresponding test objects.

    This function verifies that every object (function, class, or method) in the
    source module or class has a corresponding test object in the test module or class.

    Args:
        test_obj: The test object (module, class, or function) to check

    Raises:
        AssertionError: If any object in the source lacks a corresponding test object,
            with a detailed error message listing the untested objects

    """
    test_objs = get_objs_from_obj(test_obj)
    test_objs_paths = {make_obj_importpath(obj) for obj in test_objs}

    try:
        obj = get_obj_from_test_obj(test_obj)
    except ImportError:
        if isinstance(test_obj, ModuleType):
            # we skip if module not found bc that means it has custom tests
            # and is not part of the mirrored structure
            logger.warning("No source module found for %s, skipping", test_obj)
            return
        raise
    objs = get_objs_from_obj(obj)
    test_obj_path_to_obj = {make_test_obj_importpath_from_obj(obj): obj for obj in objs}

    missing_test_obj_path_to_obj = {
        test_path: obj
        for test_path, obj in test_obj_path_to_obj.items()
        if test_path not in test_objs_paths
    }

    # get the modules of these obj
    if missing_test_obj_path_to_obj:
        create_tests()

    msg = f"""Found missing tests. Tests skeletons were automatically created for:
    {make_untested_summary_error_msg(missing_test_obj_path_to_obj.keys())}
"""

    assert_with_msg(
        not missing_test_obj_path_to_obj,
        msg,
    )



"""Config utilities for .python-version."""

from pathlib import Path
from typing import Any

from pyrig.dev.configs.base.base import ConfigFile
from pyrig.dev.configs.pyproject import PyprojectConfigFile


class DotPythonVersionConfigFile(ConfigFile):
    """Config file for .python-version."""

    VERSION_KEY = "version"

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        return ""  # so it builds the path .python-version

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return "python-version"

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_configs(cls) -> dict[str, Any]:
        """Get the config."""
        return {
            cls.VERSION_KEY: str(
                PyprojectConfigFile.get_first_supported_python_version()
            )
        }

    @classmethod
    def load(cls) -> dict[str, Any]:
        """Load the config file."""
        return {cls.VERSION_KEY: cls.get_path().read_text(encoding="utf-8")}

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""
        if not isinstance(config, dict):
            msg = f"Cannot dump {config} to .python-version file."
            raise TypeError(msg)
        cls.get_path().write_text(config[cls.VERSION_KEY], encoding="utf-8")



"""Project utilities for introspection and manipulation.

This module provides utility functions for working with Python projects
"""

import logging
from collections.abc import Callable, Iterable
from types import ModuleType
from typing import Any

import pyrig
from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.src.modules.package import get_src_package
from pyrig.src.string import make_name_from_obj

logger = logging.getLogger(__name__)


PROJECT_MGT = "uv"

PROJECT_MGT_RUN_ARGS = [PROJECT_MGT, "run"]

RUN_PYTHON_MODULE_ARGS = ["python", "-m"]


def get_script_from_args(args: Iterable[str]) -> str:
    """Get the script from args."""
    return " ".join(args)


def get_run_python_module_args(module: ModuleType) -> list[str]:
    """Get the args to run a module."""
    from pyrig.src.modules.module import (  # noqa: PLC0415  # avoid circular import
        make_obj_importpath,
    )

    return [*RUN_PYTHON_MODULE_ARGS, make_obj_importpath(module)]


def get_project_mgt_run_module_args(module: ModuleType) -> list[str]:
    """Get the args to run a module."""
    return [*PROJECT_MGT_RUN_ARGS, *get_run_python_module_args(module)]


def get_project_mgt_run_cli_cmd_args(
    cmd: Callable[[], Any] | None = None, extra_args: list[str] | None = None
) -> list[str]:
    """Get the args to run the cli of the current project."""
    args = [
        *PROJECT_MGT_RUN_ARGS,
        PyprojectConfigFile.get_project_name_from_pkg_name(get_src_package().__name__),
    ]
    if cmd is not None:
        name = make_name_from_obj(cmd, capitalize=False)
        args.append(name)
    if extra_args is not None:
        args.extend(extra_args)
    return args


def get_project_mgt_run_pyrig_cli_cmd_args(
    cmd: Callable[[], Any] | None = None,
    extra_args: list[str] | None = None,
) -> list[str]:
    """Get the args to run pyrig."""
    args = get_project_mgt_run_cli_cmd_args(cmd, extra_args)
    args[len(PROJECT_MGT_RUN_ARGS)] = (
        PyprojectConfigFile.get_project_name_from_pkg_name(pyrig.__name__)
    )
    return args


def get_project_mgt_run_cli_cmd_script(cmd: Callable[[], Any]) -> str:
    """Get the script to run pyrig."""
    return get_script_from_args(get_project_mgt_run_cli_cmd_args(cmd))


def get_project_mgt_run_pyrig_cli_cmd_script(cmd: Callable[[], Any]) -> str:
    """Get the script to run pyrig."""
    return get_script_from_args(get_project_mgt_run_pyrig_cli_cmd_args(cmd))


def get_python_module_script(module: ModuleType) -> str:
    """Get the script to run a module."""
    return get_script_from_args(get_run_python_module_args(module))


def get_project_mgt_run_module_script(module: ModuleType) -> str:
    """Get the script to run a module."""
    return get_script_from_args(get_project_mgt_run_module_args(module))



"""Config file for function.py."""

from types import ModuleType

from pyrig.dev.configs.base.base import CopyModuleOnlyDocstringConfigFile
from pyrig.dev.tests.fixtures.scopes import function


class FunctionScopeConfigFile(CopyModuleOnlyDocstringConfigFile):
    """Config file for function.py."""

    @classmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""
        return function



"""Base class for config files."""

import inspect
from abc import ABC, abstractmethod
from pathlib import Path
from types import ModuleType
from typing import Any

import tomlkit
import yaml

import pyrig
from pyrig.dev import configs
from pyrig.src.iterate import nested_structure_is_subset
from pyrig.src.modules.class_ import (
    get_all_nonabst_subcls_from_mod_in_all_deps_depen_on_dep,
)
from pyrig.src.modules.module import (
    get_isolated_obj_name,
    get_module_content_as_str,
    make_pkg_dir,
    to_path,
)
from pyrig.src.string import split_on_uppercase
from pyrig.src.testing.convention import TESTS_PACKAGE_NAME


class ConfigFile(ABC):
    """Base class for config files."""

    @classmethod
    @abstractmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""

    @classmethod
    @abstractmethod
    def load(cls) -> dict[str, Any] | list[Any]:
        """Load the config file."""

    @classmethod
    @abstractmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""

    @classmethod
    @abstractmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""

    @classmethod
    @abstractmethod
    def get_configs(cls) -> dict[str, Any] | list[Any]:
        """Get the config."""

    def __init__(self) -> None:
        """Initialize the config file."""
        self.get_path().parent.mkdir(parents=True, exist_ok=True)
        if not self.get_path().exists():
            self.get_path().touch()
            self.dump(self.get_configs())

        if not self.is_correct():
            config = self.add_missing_configs()
            self.dump(config)

        if not self.is_correct():
            msg = f"Config file {self.get_path()} is not correct."
            raise ValueError(msg)

    @classmethod
    def get_path(cls) -> Path:
        """Get the path to the config file."""
        return (
            cls.get_parent_path() / f"{cls.get_filename()}.{cls.get_file_extension()}"
        )

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        name = cls.__name__
        abstract_parents = [
            parent.__name__ for parent in cls.__mro__ if inspect.isabstract(parent)
        ]
        for parent in abstract_parents:
            name = name.removesuffix(parent)
        return "_".join(split_on_uppercase(name)).lower()

    @classmethod
    def add_missing_configs(cls) -> dict[str, Any] | list[Any]:
        """Add any missing configs to the config file."""
        current_config = cls.load()
        expected_config = cls.get_configs()
        nested_structure_is_subset(
            expected_config,
            current_config,
            cls.add_missing_dict_val,
            cls.insert_missing_list_val,
        )
        return current_config

    @staticmethod
    def add_missing_dict_val(
        expected_dict: dict[str, Any], actual_dict: dict[str, Any], key: str
    ) -> None:
        """Add a missing dict value."""
        actual_dict[key] = expected_dict[key]

    @staticmethod
    def insert_missing_list_val(
        expected_list: list[Any], actual_list: list[Any], index: int
    ) -> None:
        """Append a missing list value."""
        actual_list.insert(index, expected_list[index])

    @classmethod
    def is_correct(cls) -> bool:
        """Check if the config is correct.

        If the file is empty, it is considered correct.
        This is so bc if a user does not want a specific config file,
        they can just make it empty and the tests will not fail.
        """
        return cls.is_unwanted() or cls.is_correct_recursively(
            cls.get_configs(), cls.load()
        )

    @classmethod
    def is_unwanted(cls) -> bool:
        """Check if the config file is unwanted.

        If the file is empty, it is considered unwanted.
        """
        return (
            cls.get_path().exists() and cls.get_path().read_text(encoding="utf-8") == ""
        )

    @staticmethod
    def is_correct_recursively(
        expected_config: dict[str, Any] | list[Any],
        actual_config: dict[str, Any] | list[Any],
    ) -> bool:
        """Check if the config is correct.

        Checks if expected is a subset recursively of actual.
        If a value is Any, it is considered correct.

        Args:
            expected_config: The expected config
            actual_config: The actual config

        Returns:
            True if the config is correct, False otherwise
        """
        return nested_structure_is_subset(expected_config, actual_config)

    @classmethod
    def get_all_subclasses(cls) -> list[type["ConfigFile"]]:
        """Get all subclasses of ConfigFile."""
        return get_all_nonabst_subcls_from_mod_in_all_deps_depen_on_dep(
            cls,
            pyrig,
            configs,
            discard_parents=True,
        )

    @classmethod
    def init_config_files(cls) -> None:
        """Initialize all subclasses."""
        cls.init_priority_config_files()

        subclasses = cls.get_all_subclasses()
        subclasses = [
            subclass
            for subclass in subclasses
            if subclass not in cls.get_priority_config_files()
        ]
        for subclass in subclasses:
            subclass()

    @classmethod
    def init_priority_config_files(cls) -> None:
        """Initialize all subclasses."""
        for subclass in cls.get_priority_config_files():
            subclass()

    @classmethod
    def get_priority_config_files(cls) -> list[type["ConfigFile"]]:
        """Get the priority config files."""
        # Some must be first:
        from pyrig.dev.configs.git.gitignore import (  # noqa: PLC0415
            GitIgnoreConfigFile,
        )
        from pyrig.dev.configs.pyproject import (  # noqa: PLC0415
            PyprojectConfigFile,
        )
        from pyrig.dev.configs.python.builder import (  # noqa: PLC0415
            BuilderConfigFile,
        )
        from pyrig.dev.configs.python.configs import (  # noqa: PLC0415
            ConfigsConfigFile,
        )
        from pyrig.dev.configs.python.main import (  # noqa: PLC0415
            MainConfigFile,
        )
        from pyrig.dev.configs.testing.conftest import (  # noqa: PLC0415
            ConftestConfigFile,
        )

        return [
            GitIgnoreConfigFile,
            PyprojectConfigFile,
            MainConfigFile,
            ConfigsConfigFile,
            BuilderConfigFile,
            ConftestConfigFile,
        ]

    @classmethod
    def get_module_name_replacing_start_module(
        cls, module: ModuleType, new_start_module_name: str
    ) -> str:
        """Get the module name of a module replacing the start module."""
        module_current_start = module.__name__.split(".")[0]
        return module.__name__.replace(module_current_start, new_start_module_name, 1)


class YamlConfigFile(ConfigFile):
    """Base class for yaml config files."""

    @classmethod
    def load(cls) -> dict[str, Any] | list[Any]:
        """Load the config file."""
        return yaml.safe_load(cls.get_path().read_text(encoding="utf-8")) or {}

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""
        with cls.get_path().open("w") as f:
            yaml.safe_dump(config, f, sort_keys=False)

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return "yaml"


class TomlConfigFile(ConfigFile):
    """Base class for toml config files."""

    @classmethod
    def load(cls) -> dict[str, Any]:
        """Load the config file."""
        return tomlkit.parse(cls.get_path().read_text(encoding="utf-8"))

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""
        if not isinstance(config, dict):
            msg = f"Cannot dump {config} to toml file."
            raise TypeError(msg)
        cls.pretty_dump(config)

    @classmethod
    def prettify_dict(cls, config: dict[str, Any]) -> dict[str, Any]:
        """Prettify a dict recursively. bx making lists multiline."""
        for key, value in config.items():
            if isinstance(value, list):
                array = tomlkit.array()
                array.multiline(multiline=True)
                for item in value:
                    array.append(item)
                config[key] = array
            elif isinstance(value, dict):
                config[key] = cls.prettify_dict(value)
        return config

    @classmethod
    def pretty_dump(cls, config: dict[str, Any]) -> None:
        """Dump the config file."""
        # trun all lists into multiline arrays
        config = cls.prettify_dict(config)
        with cls.get_path().open("w") as f:
            tomlkit.dump(config, f, sort_keys=False)

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return "toml"


class TextConfigFile(ConfigFile):
    """Base class for text config files.

    Those files just have a starting text
    and then can be added to. Like python files or README.md
    """

    CONTENT_KEY = "content"

    @classmethod
    @abstractmethod
    def get_content_str(cls) -> str:
        """Get the content."""

    @classmethod
    def load(cls) -> dict[str, str]:
        """Load the config file."""
        return {cls.CONTENT_KEY: cls.get_path().read_text(encoding="utf-8")}

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""
        if not isinstance(config, dict):
            msg = f"Cannot dump {config} to text file."
            raise TypeError(msg)
        if cls.get_file_content().strip():
            config[cls.CONTENT_KEY] = (
                config[cls.CONTENT_KEY] + "\n" + cls.get_file_content()
            )
        cls.get_path().write_text(config[cls.CONTENT_KEY], encoding="utf-8")

    @classmethod
    def get_configs(cls) -> dict[str, Any]:
        """Get the config."""
        return {cls.CONTENT_KEY: cls.get_content_str()}

    @classmethod
    def is_correct(cls) -> bool:
        """Check if the config is correct.

        Text files are correct if they exist and contain the correct content.
        """
        return (
            super().is_correct()
            or cls.get_content_str().strip() in cls.load()[cls.CONTENT_KEY]
        )

    @classmethod
    def get_file_content(cls) -> str:
        """Get the file content."""
        return cls.load()[cls.CONTENT_KEY]


class PythonConfigFile(TextConfigFile):
    """Base class for python config files."""

    CONTENT_KEY = "content"

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return "py"


class PythonPackageConfigFile(PythonConfigFile):
    """Base class for python package config files.

    They create an init file.
    """

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""
        super().dump(config)
        make_pkg_dir(cls.get_path().parent)


class CopyModuleConfigFile(PythonPackageConfigFile):
    """Config file that copies a module."""

    @classmethod
    @abstractmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file.

        Replaces the start module with the package name.
        """
        from pyrig.dev.configs.pyproject import PyprojectConfigFile  # noqa: PLC0415

        src_module = cls.get_src_module()
        new_module_name = cls.get_module_name_replacing_start_module(
            src_module, PyprojectConfigFile.get_package_name()
        )
        return to_path(new_module_name, is_package=True).parent

    @classmethod
    def get_content_str(cls) -> str:
        """Get the content."""
        src_module = cls.get_src_module()
        return get_module_content_as_str(src_module)

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        src_module = cls.get_src_module()
        return get_isolated_obj_name(src_module)


class CopyModuleOnlyDocstringConfigFile(CopyModuleConfigFile):
    """Config file that copies only the docstring of a module."""

    @classmethod
    def get_content_str(cls) -> str:
        """Get the content."""
        content = super().get_content_str()
        parts = content.split('"""', 2)
        return '"""' + parts[1] + '"""\n'


class TypedConfigFile(ConfigFile):
    """Config file for py.typed."""

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return "typed"

    @classmethod
    def load(cls) -> dict[str, Any] | list[Any]:
        """Load the config file."""
        return {}

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""
        if config:
            msg = "Cannot dump to py.typed file."
            raise ValueError(msg)

    @classmethod
    def get_configs(cls) -> dict[str, Any] | list[Any]:
        """Get the config."""
        return {}


class PythonTestsConfigFile(PythonConfigFile):
    """Base class for python config files in the tests directory."""

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path(TESTS_PACKAGE_NAME)



"""Config utilities for .env."""

from pathlib import Path
from typing import Any

from dotenv import dotenv_values

from pyrig.dev.configs.base.base import ConfigFile


class DotEnvConfigFile(ConfigFile):
    """config class for .env config files."""

    @classmethod
    def load(cls) -> dict[str, str | None]:
        """Load the config file."""
        return dotenv_values(cls.get_path())

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""
        # is not supposed to be dumped to, so just raise error
        if config:
            msg = f"Cannot dump {config} to .env file."
            raise ValueError(msg)

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return "env"

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        return ""  # so it builds the path .env and not env.env

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_configs(cls) -> dict[str, Any]:
        """Get the config."""
        return {}

    @classmethod
    def is_correct(cls) -> bool:
        """Check if the config is correct."""
        return super().is_correct() or cls.get_path().exists()



"""Class utilities for introspection and manipulation.

This module provides utility functions for working with Python classes,
including extracting methods from classes and finding classes within modules.
These utilities are particularly useful for reflection, testing,
and dynamic code generation.
"""

import inspect
from collections.abc import Callable
from importlib import import_module
from types import ModuleType
from typing import Any, overload

from pyrig.src.modules.function import is_func
from pyrig.src.modules.inspection import get_def_line, get_obj_members


def get_all_methods_from_cls(
    class_: type,
    *,
    exclude_parent_methods: bool = False,
    include_annotate: bool = False,
) -> list[Callable[..., Any]]:
    """Get all methods from a class.

    Retrieves all methods (functions or methods) from a class. Can optionally
    exclude methods inherited from parent classes.

    Args:
        class_: The class to extract methods from
        exclude_parent_methods: If True, only include methods defined in this class,
        excluding those inherited from parent classes
        include_annotate: If False, exclude __annotate__ method
        introduced in Python 3.14, defaults to False

    Returns:
        A list of callable methods from the class

    """
    from pyrig.src.modules.module import (  # noqa: PLC0415  # avoid circular import
        get_module_of_obj,
    )

    methods = [
        (method, name)
        for name, method in get_obj_members(class_, include_annotate=include_annotate)
        if is_func(method)
    ]

    if exclude_parent_methods:
        methods = [
            (method, name)
            for method, name in methods
            if get_module_of_obj(method).__name__ == class_.__module__
            and name in class_.__dict__
        ]

    only_methods = [method for method, _name in methods]
    # sort by definition order
    return sorted(only_methods, key=get_def_line)


def get_all_cls_from_module(module: ModuleType | str) -> list[type]:
    """Get all classes defined in a module.

    Retrieves all class objects that are defined directly in the specified module,
    excluding imported classes.

    Args:
        module: The module to extract classes from

    Returns:
        A list of class types defined in the module

    """
    from pyrig.src.modules.module import (  # noqa: PLC0415  # avoid circular import
        get_module_of_obj,
    )

    if isinstance(module, str):
        module = import_module(module)

    # necessary for bindings packages like AESGCM from cryptography._rust backend
    default = ModuleType("default")
    classes = [
        obj
        for _, obj in inspect.getmembers(module, inspect.isclass)
        if get_module_of_obj(obj, default).__name__ == module.__name__
    ]
    # sort by definition order
    return sorted(classes, key=get_def_line)


def get_all_subclasses(
    cls: type,
    load_package_before: ModuleType | None = None,
    *,
    discard_parents: bool = False,
) -> set[type]:
    """Get all subclasses of a class.

    Retrieves all classes that are subclasses of the specified class.
    Also returns subclasses of subclasses (recursive).

    Args:
        cls: The class to find subclasses of
        load_package_before: If provided,
            walks the package before loading the subclasses
            This is useful when the subclasses are defined in other modules that need
            to be imported before they can be found by __subclasses__
        discard_parents: If True, only keeps the most recent child class of each branch

    Returns:
        A list of subclasses of the given class

    """
    from pyrig.src.modules.package import (  # noqa: PLC0415  # avoid circular import
        walk_package,
    )

    if load_package_before:
        _ = list(walk_package(load_package_before))
    subclasses_set = {cls}
    subclasses_set.update(cls.__subclasses__())
    for subclass in cls.__subclasses__():
        subclasses_set.update(get_all_subclasses(subclass))
    if load_package_before is not None:
        # remove all not in the package
        subclasses_set = {
            subclass
            for subclass in subclasses_set
            if load_package_before.__name__ in subclass.__module__
        }
    if discard_parents:
        subclasses_set = discard_parent_classes(subclasses_set)
    return subclasses_set


def get_all_nonabstract_subclasses(
    cls: type,
    load_package_before: ModuleType | None = None,
    *,
    discard_parents: bool = False,
) -> set[type]:
    """Get all non-abstract subclasses of a class.

    Retrieves all classes that are subclasses of the specified class,
    excluding abstract classes. Also returns subclasses of subclasses
    (recursive).

    Args:
        cls: The class to find subclasses of
        load_package_before: If provided,
            walks the package before loading the subclasses
            This is useful when the subclasses are defined in other modules that need
            to be imported before they can be found by __subclasses__
        discard_parents: If True, only keeps the most recent child class of each branch

    Returns:
        A list of non-abstract subclasses of the given class

    """
    return {
        subclass
        for subclass in get_all_subclasses(
            cls,
            load_package_before=load_package_before,
            discard_parents=discard_parents,
        )
        if not inspect.isabstract(subclass)
    }


def init_all_nonabstract_subclasses(
    cls: type,
    load_package_before: ModuleType | None = None,
    *,
    discard_parents: bool = False,
) -> None:
    """Initialize all non-abstract subclasses of a class.

    Args:
        cls: The class to find subclasses of
        load_package_before: If provided,
            walks the package before loading the subclasses
            This is useful when the subclasses are defined in other modules that need
            to be imported before they can be found by __subclasses__
        discard_parents: If True, only keeps the most recent child class of each branch

    """
    for subclass in get_all_nonabstract_subclasses(
        cls,
        load_package_before=load_package_before,
        discard_parents=discard_parents,
    ):
        subclass()


def get_all_nonabst_subcls_from_mod_in_all_deps_depen_on_dep(
    cls: type,
    dep: ModuleType,
    load_package_before: ModuleType,
    *,
    discard_parents: bool = False,
) -> list[type]:
    """Get all non-abstract subclasses of a class from a module in all deps.

    Retrieves all classes that are subclasses of the specified class,
    excluding abstract classes. Also returns subclasses of subclasses
    (recursive).

    Args:
        cls: The class to find subclasses of
        dep: The dependency to find subclasses of
        load_package_before: If provided,
            walks the package before loading the subclasses
            This is useful when the subclasses are defined in other modules that need
            to be imported before they can be found by __subclasses__
        discard_parents: If True, only keeps the most recent child class of each branch

    Returns:
        A list of non-abstract subclasses of the given class
        Order is garanteed only that classes from the same module are grouped together

    """
    from pyrig.src.modules.module import (  # noqa: PLC0415  # avoid circular import
        get_same_modules_from_deps_depen_on_dep,
    )

    subclasses: list[type] = []
    for pkg in get_same_modules_from_deps_depen_on_dep(load_package_before, dep):
        subclasses.extend(
            get_all_nonabstract_subclasses(
                cls,
                load_package_before=pkg,
                discard_parents=discard_parents,
            )
        )
    # as these are different modules and pks we need to discard parents again
    if discard_parents:
        subclasses = discard_parent_classes(subclasses)
    return subclasses


@overload
def discard_parent_classes(classes: list[type]) -> list[type]: ...


@overload
def discard_parent_classes(classes: set[type]) -> set[type]: ...


def discard_parent_classes(classes: list[type] | set[type]) -> list[type] | set[type]:
    """Discard parent classes from a list of classes.

    Args:
        classes: The classes to discard parent classes from

    Returns:
        A list of classes with parent classes discarded

    """
    for cls in classes.copy():
        if any(child in classes for child in cls.__subclasses__()):
            classes.remove(cls)
    return classes



"""Adds the src folder with an init file."""

from pathlib import Path
from types import ModuleType

from pyrig import src
from pyrig.dev.configs.base.base import CopyModuleOnlyDocstringConfigFile
from pyrig.src.modules.module import get_isolated_obj_name


class SrcInitConfigFile(CopyModuleOnlyDocstringConfigFile):
    """Config file for src/__init__.py."""

    @classmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""
        return src

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        path = super().get_parent_path()
        # this path will be parent of src
        return path / get_isolated_obj_name(src)

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        return "__init__"



"""Pyrigs init script.

This script inits the project by calling all setup steps.
"""

import logging
from collections.abc import Callable
from typing import Any

from pyrig.dev.configs.base.base import ConfigFile
from pyrig.dev.configs.git.pre_commit import PreCommitConfigConfigFile
from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.dev.configs.testing.conftest import ConftestConfigFile
from pyrig.src.project.create_root import create_root
from pyrig.src.testing.create_tests import create_tests

logger = logging.getLogger(__name__)


SETUP_STEPS: list[Callable[..., Any]] = [
    ConfigFile.init_priority_config_files,  # write dev deps to pyproject.toml
    PyprojectConfigFile.install_dependencies,  # to install dev deps
    PyprojectConfigFile.update_dependencies,  # to update dev deps
    create_root,
    create_tests,
    PreCommitConfigConfigFile.run_hooks,
    ConftestConfigFile.run_tests,
    PyprojectConfigFile.install_dependencies,  # to activate cli
]


def init() -> None:
    """Set up the project."""
    for step in SETUP_STEPS:
        logger.info("Running setup step: %s", step.__name__)
        step()
    logger.info("Setup complete!")



"""OS utilities for finding commands and paths.

This module provides utility functions for working with the operating system,
including finding the path to commands and managing environment variables.
These utilities help with system-level operations and configuration.
"""

import logging
import shutil
import subprocess  # nosec: B404
from collections.abc import Sequence
from pathlib import Path
from typing import Any

logger = logging.getLogger(__name__)


def which_with_raise(cmd: str, *, raise_error: bool = True) -> str | None:
    """Give the path to the given command.

    Args:
        cmd: The command to find
        raise_error: Whether to raise an error if the command is not found

    Returns:
        The path to the command

    Raises:
        FileNotFoundError: If the command is not found

    """
    path = shutil.which(cmd)
    if path is None:
        msg = f"Command {cmd} not found"
        if raise_error:
            raise FileNotFoundError(msg)
    return path


def run_subprocess(  # noqa: PLR0913
    args: Sequence[str],
    *,
    input_: str | bytes | None = None,
    capture_output: bool = True,
    timeout: int | None = None,
    check: bool = True,
    cwd: str | Path | None = None,
    **kwargs: Any,
) -> subprocess.CompletedProcess[Any]:
    """Run a subprocess.

    Args:
        args: The arguments to pass to the subprocess
        input_: The input to pass to the subprocess
        capture_output: Whether to capture the output of the subprocess
        timeout: The timeout for the subprocess
        check: to raise an exception if the subprocess returns a non-zero exit code
        cwd: The working directory to run the subprocess in
        kwargs: Any other arguments to pass to subprocess.run()

    """
    if cwd is None:
        cwd = Path.cwd()
    try:
        return subprocess.run(  # noqa: S603  # nosec: B603
            args,
            check=check,
            input=input_,
            capture_output=capture_output,
            timeout=timeout,
            cwd=cwd,
            **kwargs,
        )
    except subprocess.CalledProcessError as e:
        logger.exception(
            """
Failed to run subprocess:
    args: %s
    returncode: %s
    stdout: %s
    stderr: %s
""",
            args,
            e.returncode,
            e.stdout.decode("utf-8"),
            e.stderr.decode("utf-8"),
        )
        raise



"""Config utilities for conftest.py."""

from subprocess import CompletedProcess  # nosec: B404

from pyrig.dev.configs.base.base import PythonTestsConfigFile
from pyrig.src.modules.module import make_obj_importpath
from pyrig.src.os.os import run_subprocess
from pyrig.src.project.mgt import PROJECT_MGT_RUN_ARGS


class ConftestConfigFile(PythonTestsConfigFile):
    """Config file for conftest.py."""

    @classmethod
    def get_content_str(cls) -> str:
        """Get the config content."""
        from pyrig.dev.tests import conftest  # noqa: PLC0415

        return f'''"""Pytest configuration for tests.

This module configures pytest plugins for the test suite, setting up the necessary
fixtures and hooks for the different
test scopes (function, class, module, package, session).
It also import custom plugins from tests/base/scopes.
This file should not be modified manually.
"""

pytest_plugins = ["{make_obj_importpath(conftest)}"]
'''

    @classmethod
    def run_tests(cls, *, check: bool = True) -> CompletedProcess[str]:
        """Run the tests."""
        return run_subprocess([*PROJECT_MGT_RUN_ARGS, "pytest"], check=check)



"""Contains the pull request workflow.

This workflow is used to run tests on pull requests.
"""

from typing import Any

from pyrig.dev.configs.workflows.base.base import Workflow


class HealthCheckWorkflow(Workflow):
    """Pull request workflow.

    This workflow is triggered by a pull request.
    It runs tests on the pull request.
    """

    @classmethod
    def get_workflow_triggers(cls) -> dict[str, Any]:
        """Get the workflow triggers."""
        triggers = super().get_workflow_triggers()
        triggers.update(cls.on_pull_request())
        triggers.update(cls.on_schedule(cron="0 6 * * *"))
        return triggers

    @classmethod
    def get_jobs(cls) -> dict[str, Any]:
        """Get the workflow jobs."""
        jobs: dict[str, Any] = {}
        jobs.update(cls.job_health_check_matrix())
        jobs.update(cls.job_health_check())
        return jobs

    @classmethod
    def job_health_check_matrix(cls) -> dict[str, Any]:
        """Get the health check matrix job."""
        return cls.get_job(
            job_func=cls.job_health_check_matrix,
            strategy=cls.strategy_matrix_os_and_python_version(),
            runs_on=cls.insert_matrix_os(),
            steps=cls.steps_health_check_matrix(),
        )

    @classmethod
    def job_health_check(cls) -> dict[str, Any]:
        """Get the health check job."""
        return cls.get_job(
            job_func=cls.job_health_check,
            needs=[cls.make_id_from_func(cls.job_health_check_matrix)],
            steps=cls.steps_aggregate_matrix_results(),
        )

    @classmethod
    def steps_health_check_matrix(cls) -> list[dict[str, Any]]:
        """Get the health check matrix steps."""
        return [
            *cls.steps_core_matrix_setup(
                python_version=cls.insert_matrix_python_version()
            ),
            cls.step_protect_repository(),
            cls.step_run_pre_commit_hooks(),
            cls.step_run_tests(),
        ]

    @classmethod
    def steps_aggregate_matrix_results(cls) -> list[dict[str, Any]]:
        """Get the aggregate matrix results step."""
        return [
            cls.step_aggregate_matrix_results(),
        ]



"""Script to protect the repo and branches of a repository."""

from typing import Any

from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.src.git.github.github import (
    get_github_repo_token,
    get_repo_owner_and_name_from_git,
)
from pyrig.src.git.github.repo.repo import (
    DEFAULT_BRANCH,
    DEFAULT_RULESET_NAME,
    create_or_update_ruleset,
    get_repo,
    get_rules_payload,
)


def protect_repository() -> None:
    """Protect the repository."""
    set_secure_repo_settings()
    create_or_update_default_branch_ruleset()


def set_secure_repo_settings() -> None:
    """Set standard settings for the repository."""
    owner, repo_name = get_repo_owner_and_name_from_git()
    token = get_github_repo_token()
    repo = get_repo(token, owner, repo_name)

    toml_description = PyprojectConfigFile.get_project_description()

    repo.edit(
        name=repo_name,
        description=toml_description,
        default_branch=DEFAULT_BRANCH,
        delete_branch_on_merge=True,
        allow_update_branch=True,
        allow_merge_commit=False,
        allow_rebase_merge=True,
        allow_squash_merge=True,
    )


def create_or_update_default_branch_ruleset() -> None:
    """Add a branch protection rule to the repository."""
    create_or_update_ruleset(
        **get_default_ruleset_params(),
    )


def get_default_ruleset_params() -> dict[str, Any]:
    """Get the default ruleset parameters."""
    from pyrig.dev.configs.workflows.health_check import (  # noqa: PLC0415
        HealthCheckWorkflow,  # avoid circular import
    )

    owner, repo_name = get_repo_owner_and_name_from_git()
    token = get_github_repo_token()

    rules = get_rules_payload(
        deletion={},
        non_fast_forward={},
        creation={},
        update={},
        pull_request={
            "required_approving_review_count": 1,
            "dismiss_stale_reviews_on_push": True,
            "require_code_owner_review": True,
            "require_last_push_approval": True,
            "required_review_thread_resolution": True,
            "allowed_merge_methods": ["squash", "rebase"],
        },
        required_linear_history={},
        required_signatures={},
        required_status_checks={
            "strict_required_status_checks_policy": True,
            "do_not_enforce_on_create": False,
            "required_status_checks": [
                {
                    "context": HealthCheckWorkflow.get_filename(),
                }
            ],
        },
    )

    return {
        "owner": owner,
        "token": token,
        "repo_name": repo_name,
        "ruleset_name": DEFAULT_RULESET_NAME,
        "enforcement": "active",
        "bypass_actors": [
            {
                "actor_id": 5,
                "actor_type": "RepositoryRole",
                "bypass_mode": "always",
            }
        ],
        "target": "branch",
        "conditions": {"ref_name": {"include": ["~DEFAULT_BRANCH"], "exclude": []}},
        "rules": rules,
    }



"""Utilities for automatically creating test files for the project.

This module provides functions to generate test files for all modules and classes
in the project, ensuring that every function and method has a corresponding test.
It creates the basic test structure and generates skeleton test functions with
NotImplementedError to indicate tests that need to be written.
"""

from types import ModuleType
from typing import cast

from pyrig.src.modules.class_ import (
    get_all_cls_from_module,
    get_all_methods_from_cls,
)
from pyrig.src.modules.function import get_all_functions_from_module
from pyrig.src.modules.inspection import get_qualname_of_obj
from pyrig.src.modules.module import (
    create_module,
    get_isolated_obj_name,
    get_module_content_as_str,
    to_path,
)
from pyrig.src.modules.package import (
    get_src_package,
    walk_package,
)
from pyrig.src.testing.convention import (
    get_test_obj_from_obj,
    make_test_obj_importpath_from_obj,
    make_test_obj_name,
)


def create_tests() -> None:
    """Create all test files for the project.

    This function orchestrates the test creation process by first setting up the base
    test structure and then creating test files for all source packages.
    """
    create_tests_for_package(get_src_package())


def create_tests_for_package(package: ModuleType) -> None:
    """Create test files for all modules in the source package.

    This function walks through the source package hierarchy and creates corresponding
    test packages and modules for each package and module found in the source.
    """
    for pkg, modules in walk_package(package):
        create_test_package(pkg)
        for module in modules:
            create_test_module(module)


def create_test_package(package: ModuleType) -> None:
    """Create a test package for a source package.

    Args:
        package: The source package module to create a test package for

    This function creates a test package with the appropriate naming convention
    if it doesn't already exist.

    """
    test_package_name = make_test_obj_importpath_from_obj(package)
    # create package if it doesn't exist
    create_module(test_package_name, is_package=True)


def create_test_module(module: ModuleType) -> None:
    """Create a test module for a source module.

    Args:
        module: The source module to create a test module for

    This function:
    1. Creates a test module with the appropriate naming convention
    2. Generates the test module content with skeleton test functions
    3. Writes the content to the test module file

    """
    test_module_name = make_test_obj_importpath_from_obj(module)
    test_module = create_module(test_module_name, is_package=False)
    test_module_path = to_path(test_module, is_package=False)
    test_module_path.write_text(get_test_module_content(module))


def get_test_module_content(module: ModuleType) -> str:
    """Generate the content for a test module.

    Args:
        module: The source module to generate test content for

    Returns:
        The generated test module content as a string

    This function:
    1. Gets the existing test module content if it exists
    2. Adds test functions for all functions in the source module
    3. Adds test classes for all classes in the source module

    """
    test_module = cast("ModuleType", (get_test_obj_from_obj(module)))
    test_module_content = get_module_content_as_str(test_module)

    test_module_content = get_test_functions_content(
        module, test_module, test_module_content
    )

    return get_test_classes_content(module, test_module, test_module_content)


def get_test_functions_content(
    module: ModuleType,
    test_module: ModuleType,
    test_module_content: str,
) -> str:
    """Generate test function content for a module.

    Args:
        module: The source module containing functions to test
        test_module: The test module to add function tests to
        test_module_content: The current content of the test module

    Returns:
        The updated test module content with function tests added

    This function:
    1. Identifies all functions in the source module
    2. Determines which functions don't have corresponding tests
    3. Generates skeleton test functions for untested functions

    """
    funcs = get_all_functions_from_module(module)
    test_functions = get_all_functions_from_module(test_module)
    supposed_test_funcs_names = [make_test_obj_name(f) for f in funcs]

    test_funcs_names = [get_qualname_of_obj(f) for f in test_functions]

    untested_funcs_names = [
        f for f in supposed_test_funcs_names if f not in test_funcs_names
    ]

    for test_func_name in untested_funcs_names:
        test_module_content += f"""

def {test_func_name}() -> None:
    \"\"\"Test function.\"\"\"
    raise {NotImplementedError.__name__}
"""

    return test_module_content


def get_test_classes_content(
    module: ModuleType,
    test_module: ModuleType,
    test_module_content: str,
) -> str:
    """Generate test class content for a module.

    Args:
        module: The source module containing classes to test
        test_module: The test module to add class tests to
        test_module_content: The current content of the test module

    Returns:
        The updated test module content with class tests added

    This function:
    1. Identifies all classes in the source module
    2. Determines which classes and methods don't have corresponding tests
    3. Generates skeleton test classes and methods for untested classes and methods
    4. Inserts the new test classes into the existing content
       if the class already exists

    Raises:
        ValueError: If a test class declaration appears multiple
                    times in the test module

    """
    classes = get_all_cls_from_module(module)
    test_classes = get_all_cls_from_module(test_module)

    class_to_methods = {
        c: get_all_methods_from_cls(c, exclude_parent_methods=True) for c in classes
    }
    test_class_to_methods = {
        tc: get_all_methods_from_cls(tc, exclude_parent_methods=True)
        for tc in test_classes
    }

    supposed_test_class_to_methods_names = {
        make_test_obj_name(c): [make_test_obj_name(m) for m in ms]
        for c, ms in class_to_methods.items()
    }
    test_class_to_methods_names = {
        get_isolated_obj_name(tc): [get_isolated_obj_name(tm) for tm in tms]
        for tc, tms in test_class_to_methods.items()
    }

    untested_test_class_to_methods_names: dict[str, list[str]] = {}
    for (
        test_class_name,
        supposed_test_methods_names,
    ) in supposed_test_class_to_methods_names.items():
        test_methods_names = test_class_to_methods_names.get(test_class_name, [])
        untested_methods_names = [
            tmn for tmn in supposed_test_methods_names if tmn not in test_methods_names
        ]
        if (
            not supposed_test_methods_names
            and test_class_name not in test_class_to_methods_names
        ):
            untested_test_class_to_methods_names[test_class_name] = []
        if untested_methods_names:
            untested_test_class_to_methods_names[test_class_name] = (
                untested_methods_names
            )

    for (
        test_class_name,
        untested_methods_names,
    ) in untested_test_class_to_methods_names.items():
        test_class_declaration = f"""
class {test_class_name}:
    \"\"\"Test class.\"\"\"
"""
        test_class_content = test_class_declaration
        for untested_method_name in untested_methods_names:
            test_class_content += f"""
    def {untested_method_name}(self) -> None:
        \"\"\"Test method.\"\"\"
        raise {NotImplementedError.__name__}
"""
        parts = test_module_content.split(test_class_declaration)
        expected_parts = 2
        if len(parts) > expected_parts:
            msg = f"Found {len(parts)} parts, expected 2"
            raise ValueError(msg)
        parts.insert(1, test_class_content)
        test_module_content = "".join(parts)

    return test_module_content



"""Config utilities for .gitignore."""

import os
from pathlib import Path
from typing import Any

import pathspec
import requests

import pyrig
from pyrig.dev.configs.base.base import ConfigFile
from pyrig.dev.configs.dot_env import DotEnvConfigFile
from pyrig.dev.configs.python.experiment import ExperimentConfigFile


class GitIgnoreConfigFile(ConfigFile):
    """Config file for .gitignore."""

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        return ""  # so it builds the path .gitignore and not gitignore.gitignore

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return "gitignore"

    @classmethod
    def load(cls) -> list[str]:
        """Load the config file."""
        return cls.get_path().read_text(encoding="utf-8").splitlines()

    @classmethod
    def dump(cls, config: list[str] | dict[str, Any]) -> None:
        """Dump the config file."""
        if not isinstance(config, list):
            msg = f"Cannot dump {config} to .gitignore file."
            raise TypeError(msg)
        cls.get_path().write_text("\n".join(config), encoding="utf-8")

    @classmethod
    def get_configs(cls) -> list[str]:
        """Get the config."""
        # fetch the standard github gitignore via https://github.com/github/gitignore/blob/main/Python.gitignore
        needed = [
            *cls.get_github_python_gitignore(),
            "# vscode stuff",
            ".vscode/",
            "",
            f"# {pyrig.__name__} stuff",
            "# for walk_os_skipping_gitignore_patterns func",
            ".git/",
            "# for executing experimental code",
            "/" + ExperimentConfigFile.get_path().as_posix(),
        ]

        dotenv_path = DotEnvConfigFile.get_path().as_posix()
        if dotenv_path not in needed:
            needed.extend(["# for secrets used locally", dotenv_path])

        existing = cls.load()
        needed = [p for p in needed if p not in set(existing)]
        return existing + needed

    @classmethod
    def get_github_python_gitignore(cls) -> list[str]:
        """Get the standard github python gitignore."""
        url = "https://raw.githubusercontent.com/github/gitignore/main/Python.gitignore"
        res = requests.get(url, timeout=10)
        if not res.ok:
            if not Path(".gitignore").exists():
                msg = f"Failed to fetch {url}. Cannot create .gitignore."
                raise RuntimeError(msg)
            return []
        return res.text.splitlines()

    @classmethod
    def path_is_in_gitignore(cls, relative_path: str | Path) -> bool:
        """Check if a path matches any pattern in the .gitignore file.

        Args:
            relative_path: The path to check, relative to the repository root

        Returns:
            True if the path matches any pattern in .gitignore, False otherwise

        """
        gitignore_path = cls.get_path()
        if not gitignore_path.exists():
            return False
        as_path = Path(relative_path)
        if as_path.is_absolute():
            as_path = as_path.relative_to(Path.cwd())
        is_dir = (
            bool(as_path.suffix == "")
            or as_path.is_dir()
            or str(as_path).endswith(os.sep)
        )
        is_dir = is_dir and not as_path.is_file()

        as_posix = as_path.as_posix()
        if is_dir and not as_posix.endswith("/"):
            as_posix += "/"

        spec = pathspec.PathSpec.from_lines(
            "gitwildmatch",
            cls.load(),
        )

        return spec.match_file(as_posix)



"""Pytest configuration for pirig tests.

finds all the plugins in the tests directory and the package's testing module
and adds them to pytest_plugins. This way defining reusable fixtures is easy.
"""

from pathlib import Path

import pyrig
from pyrig.dev.tests import fixtures
from pyrig.src.modules.module import (
    get_same_modules_from_deps_depen_on_dep,
    to_module_name,
    to_path,
)

# find the fixtures module in all packages that depend on pyrig
# and add all paths to pytest_plugins
fixtures_pkgs = get_same_modules_from_deps_depen_on_dep(fixtures, pyrig)


pytest_plugin_paths: list[Path] = []
for pkg in fixtures_pkgs:
    absolute_path = Path(pkg.__path__[0])
    relative_path = to_path(pkg.__name__, is_package=True)

    pkg_root = Path(absolute_path.as_posix().removesuffix(relative_path.as_posix()))

    for path in absolute_path.rglob("*.py"):
        rel_plugin_path = path.relative_to(pkg_root)
        pytest_plugin_paths.append(rel_plugin_path)

pytest_plugins = [to_module_name(path) for path in pytest_plugin_paths]



"""Has utils towards skipping tests."""

import functools

import pytest

from pyrig.src.git.github.github import running_in_github_actions

skip_fixture_test: pytest.MarkDecorator = functools.partial(
    pytest.mark.skip,
    reason="Fixtures are not testable bc they cannot be called directly.",
)()


skip_in_github_actions: pytest.MarkDecorator = functools.partial(
    pytest.mark.skipif,
    running_in_github_actions(),
    reason="Test cannot run in GitHub action.",
)()



"""Config utilities for subcommands.py."""

from types import ModuleType

from pyrig.dev.cli import subcommands
from pyrig.dev.configs.base.base import CopyModuleOnlyDocstringConfigFile


class SubcommandsConfigFile(CopyModuleOnlyDocstringConfigFile):
    """Config file for subcommands.py."""

    @classmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""
        return subcommands



"""Inspection utilities for introspecting Python objects.

This module provides utility functions for inspecting Python objects,
including checking if an object is a function or method, and unwrapping
methods to their underlying functions.
"""

import inspect
import sys
from collections.abc import Callable
from typing import Any, cast


def get_obj_members(
    obj: Any, *, include_annotate: bool = False
) -> list[tuple[str, Any]]:
    """Get all members of an object."""
    members = [(member, value) for member, value in inspect.getmembers(obj)]
    if not include_annotate:
        members = [
            (member, value)
            for member, value in members
            if member not in ("__annotate__", "__annotate_func__")
        ]
    return members


def inside_frozen_bundle() -> bool:
    """Return True if the code is running inside a frozen bundle."""
    return getattr(sys, "frozen", False)


def get_def_line(obj: Any) -> int:
    """Return the line number where a method-like object is defined."""
    if isinstance(obj, property):
        obj = obj.fget
    unwrapped = inspect.unwrap(obj)
    if hasattr(unwrapped, "__code__"):
        return int(unwrapped.__code__.co_firstlineno)
    # getsourcelines does not work if in a pyinstaller bundle or something
    if inside_frozen_bundle():
        return 0
    return inspect.getsourcelines(unwrapped)[1]


def get_unwrapped_obj(obj: Any) -> Any:
    """Return the unwrapped version of a method-like object."""
    if isinstance(obj, property):
        obj = obj.fget  # get the getter function of the property
    return inspect.unwrap(obj)


def get_qualname_of_obj(obj: Callable[..., Any] | type) -> str:
    """Return the name of a method-like object."""
    unwrapped = get_unwrapped_obj(obj)
    return cast("str", unwrapped.__qualname__)



"""Contains the publish workflow.

This workflow is used to publish the package to PyPI.
"""

from typing import Any

from pyrig.dev.configs.workflows.base.base import Workflow
from pyrig.dev.configs.workflows.release import ReleaseWorkflow


class PublishWorkflow(Workflow):
    """Publish workflow.

    This workflow is triggered by the release workflow.
    It publishes the package to PyPI.
    """

    @classmethod
    def get_workflow_triggers(cls) -> dict[str, Any]:
        """Get the workflow triggers."""
        triggers = super().get_workflow_triggers()
        triggers.update(
            cls.on_workflow_run(workflows=[ReleaseWorkflow.get_workflow_name()])
        )
        return triggers

    @classmethod
    def get_jobs(cls) -> dict[str, Any]:
        """Get the workflow jobs."""
        jobs: dict[str, Any] = {}
        jobs.update(cls.job_publish())
        return jobs

    @classmethod
    def job_publish(cls) -> dict[str, Any]:
        """Get the publish job."""
        return cls.get_job(
            job_func=cls.job_publish,
            steps=cls.steps_publish(),
            if_condition=cls.if_workflow_run_is_success(),
        )

    @classmethod
    def steps_publish(cls) -> list[dict[str, Any]]:
        """Get the publish steps."""
        return [
            *cls.steps_core_setup(),
            cls.step_build_wheel(),
            cls.step_publish_to_pypi(),
        ]



"""This module contains the CLI entrypoint."""

import sys
from pathlib import Path

import typer

from pyrig import main as pyrig_main
from pyrig.dev.cli import subcommands
from pyrig.dev.configs.base.base import ConfigFile
from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.src.modules.function import get_all_functions_from_module
from pyrig.src.modules.module import import_module_from_path

app = typer.Typer()


def add_subcommands() -> None:
    """Add subcommands to the CLI."""
    # extract project name from sys.argv[0]
    project_name = Path(sys.argv[0]).name
    pkg_name = PyprojectConfigFile.get_pkg_name_from_project_name(project_name)

    main_module_name = PyprojectConfigFile.get_module_name_replacing_start_module(
        pyrig_main, pkg_name
    )
    main_module = import_module_from_path(main_module_name)
    app.command()(main_module.main)

    # replace the first parent with pkg_name
    subcommands_module_name = ConfigFile.get_module_name_replacing_start_module(
        subcommands, pkg_name
    )

    subcommands_module = import_module_from_path(subcommands_module_name)

    sub_cmds = get_all_functions_from_module(subcommands_module)

    for sub_cmd in sub_cmds:
        app.command()(sub_cmd)


def main() -> None:
    """Entry point for the CLI."""
    add_subcommands()
    app()



"""Config utilities for test_zero.py."""

from pyrig.dev.configs.base.base import PythonTestsConfigFile
from pyrig.src.os.os import run_subprocess
from pyrig.src.project.mgt import get_project_mgt_run_module_args
from pyrig.src.testing import create_tests


class ZeroTestConfigFile(PythonTestsConfigFile):
    """Config file for test_zero.py."""

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        filename = super().get_filename()
        return "_".join(reversed(filename.split("_")))

    @classmethod
    def get_content_str(cls) -> str:
        """Get the config."""
        return '''"""Contains an empty test."""


def test_zero() -> None:
    """Empty test.

    Exists so that when no tests are written yet the base fixtures are executed.
    """
'''

    @classmethod
    def create_tests(cls) -> None:
        """Create the tests."""
        run_subprocess(get_project_mgt_run_module_args(create_tests))



"""Config utilities for pyproject.toml."""

import re
from functools import cache
from pathlib import Path
from subprocess import CompletedProcess  # nosec: B404
from typing import Any

import requests
from packaging.version import Version

from pyrig.dev.configs.base.base import TomlConfigFile
from pyrig.dev.configs.python.experiment import ExperimentConfigFile
from pyrig.src.os.os import run_subprocess
from pyrig.src.project.versions import VersionConstraint
from pyrig.src.testing.convention import TEST_MODULE_PREFIX, TESTS_PACKAGE_NAME


class PyprojectConfigFile(TomlConfigFile):
    """Config file for pyproject.toml."""

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file."""
        if not isinstance(config, dict):
            msg = f"Cannot dump {config} to pyproject.toml file."
            raise TypeError(msg)
        # remove the versions from the dependencies
        cls.remove_wrong_dependencies(config)
        super().dump(config)

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_project_name_from_cwd(cls) -> str:
        """Get the repository name.

        Is the parent folder the project ives in and should be the same as the
        project name.
        """
        cwd = Path.cwd()
        return cwd.name

    @classmethod
    def get_pkg_name_from_cwd(cls) -> str:
        """Get the package name from the cwd."""
        return cls.get_pkg_name_from_project_name(cls.get_project_name_from_cwd())

    @classmethod
    def get_project_description(cls) -> str:
        """Get the project description."""
        return str(cls.load().get("project", {}).get("description", ""))

    @classmethod
    def get_configs(cls) -> dict[str, Any]:
        """Get the config."""
        from pyrig.dev.cli import (  # noqa: PLC0415
            cli,
        )

        return {
            "project": {
                "name": cls.get_project_name_from_cwd(),
                "readme": "README.md",
                "scripts": {
                    cls.get_project_name(): f"{cli.__name__}:{cli.main.__name__}"
                },
                "dependencies": cls.make_dependency_versions(cls.get_dependencies()),
            },
            "dependency-groups": {
                "dev": cls.make_dependency_versions(
                    cls.get_dev_dependencies(),
                    additional=cls.get_standard_dev_dependencies(),
                )
            },
            "build-system": {
                "requires": ["uv_build"],
                "build-backend": "uv_build",
            },
            "tool": {
                "uv": {
                    "build-backend": {
                        "module-name": cls.get_pkg_name_from_cwd(),
                        "module-root": "",
                    }
                },
                "ruff": {
                    "exclude": [".*", "**/migrations/*.py"],
                    "lint": {
                        "select": ["ALL"],
                        "ignore": ["D203", "D213", "COM812", "ANN401"],
                        "fixable": ["ALL"],
                        "per-file-ignores": {
                            f"{TESTS_PACKAGE_NAME}/**/*.py": ["S101"],
                        },
                        "pydocstyle": {"convention": "google"},
                    },
                },
                "mypy": {
                    "strict": True,
                    "warn_unreachable": True,
                    "show_error_codes": True,
                    "files": ".",
                },
                "pytest": {"ini_options": {"testpaths": [TESTS_PACKAGE_NAME]}},
                "bandit": {
                    "exclude_dirs": [
                        "./" + ExperimentConfigFile.get_path().as_posix(),
                        ".*",
                    ],
                    "assert_used": {
                        "skips": [f"*{TEST_MODULE_PREFIX}*.py"],
                    },
                },
            },
        }

    @classmethod
    def should_remove_version_from_dep(cls) -> bool:
        """Check if we should remove the version from the dependency.

        We should remove the version if we are in a dev dependency and the dep
        is in the standard dev dependencies.
        Can be overridden by subclasses.
        """
        return True

    @classmethod
    def remove_wrong_dependencies(cls, config: dict[str, Any]) -> None:
        """Remove the wrong dependencies from the config."""
        # removes the versions from the dependencies
        config["project"]["dependencies"] = cls.make_dependency_versions(
            config["project"]["dependencies"]
        )
        config["dependency-groups"]["dev"] = cls.make_dependency_versions(
            config["dependency-groups"]["dev"]
        )

    @classmethod
    def make_dependency_versions(
        cls,
        dependencies: list[str],
        additional: list[str] | None = None,
    ) -> list[str]:
        """Make a dependency to version dict.

        Args:
            dependencies: Dependencies to add
            additional: Additional dependencies to add

        Returns:
            Dependency to version dict
        """
        if additional is None:
            additional = []
        dependencies.extend(additional)
        deps: list[str] = []
        for dep in dependencies:
            at_file_dep = "file://"
            if at_file_dep in dep:
                new_dep = dep
            elif cls.should_remove_version_from_dep():
                # remove version if it exists by split re on first non alnum or _ -
                new_dep = cls.remove_version_from_dep(dep)
            else:
                new_dep = dep
            deps.append(new_dep)
        return sorted(set(deps))

    @classmethod
    def remove_version_from_dep(cls, dep: str) -> str:
        """Remove the version from a dependency."""
        return re.split(r"[^a-zA-Z0-9_-]", dep)[0]

    @classmethod
    def get_package_name(cls) -> str:
        """Get the package name."""
        project_name = cls.get_project_name()
        return cls.get_pkg_name_from_project_name(project_name)

    @classmethod
    def get_pkg_name_from_project_name(cls, project_name: str) -> str:
        """Get the package name from the project name."""
        return project_name.replace("-", "_")

    @classmethod
    def get_project_name_from_pkg_name(cls, pkg_name: str) -> str:
        """Get the project name from the package name."""
        return pkg_name.replace("_", "-")

    @classmethod
    def get_project_name(cls) -> str:
        """Get the project name."""
        return str(cls.load().get("project", {}).get("name", ""))

    @classmethod
    def get_all_dependencies(cls) -> list[str]:
        """Get all dependencies."""
        all_deps = cls.get_dependencies()
        all_deps.extend(cls.get_dev_dependencies())
        return all_deps

    @classmethod
    def get_standard_dev_dependencies(cls) -> list[str]:
        """Get the standard dev dependencies."""
        standard_dev_dependencies: list[str] = [
            "bandit",
            "mypy",
            "pre-commit",
            "pytest",
            "pytest-mock",
            "ruff",
            "types-defusedxml",
            "types-networkx",
            "types-pyinstaller",
            "types-pyyaml",
            "types-setuptools",
            "types-tqdm",
            "pyinstaller",
        ]
        # sort the dependencies
        return sorted(standard_dev_dependencies)

    @classmethod
    def get_dev_dependencies(cls) -> list[str]:
        """Get the dev dependencies."""
        dev_deps: list[str] = cls.load().get("dependency-groups", {}).get("dev", [])
        return dev_deps

    @classmethod
    def get_dependencies(cls) -> list[str]:
        """Get the dependencies."""
        deps: list[str] = cls.load().get("project", {}).get("dependencies", [])
        return deps

    @classmethod
    @cache
    def fetch_latest_python_version(cls) -> Version:
        """Fetch the latest python version from python.org."""
        url = "https://endoflife.date/api/python.json"
        resp = requests.get(url, timeout=10)
        resp.raise_for_status()
        data = resp.json()
        # first element has metadata for latest stable
        latest_version = data[0]["latest"]
        return Version(latest_version)

    @classmethod
    def get_latest_possible_python_version(cls) -> Version:
        """Get the latest possible python version."""
        constraint = cls.load()["project"]["requires-python"]
        version_constraint = VersionConstraint(constraint)
        version = version_constraint.get_upper_inclusive()
        if version is None:
            version = cls.fetch_latest_python_version()
        return version

    @classmethod
    def get_first_supported_python_version(cls) -> Version:
        """Get the first supported python version."""
        constraint = cls.load()["project"]["requires-python"]
        version_constraint = VersionConstraint(constraint)
        lower = version_constraint.get_lower_inclusive()
        if lower is None:
            msg = "Need a lower bound for python version"
            raise ValueError(msg)
        return lower

    @classmethod
    def get_supported_python_versions(cls) -> list[Version]:
        """Get all supported python versions."""
        constraint = cls.load()["project"]["requires-python"]
        version_constraint = VersionConstraint(constraint)
        return version_constraint.get_version_range(
            level="minor", upper_default=cls.fetch_latest_python_version()
        )

    @classmethod
    def update_dependencies(cls, *, check: bool = True) -> CompletedProcess[bytes]:
        """Update the dependencies."""
        from pyrig.src.project.mgt import PROJECT_MGT  # noqa: PLC0415

        return run_subprocess([PROJECT_MGT, "lock", "--upgrade"], check=check)

    @classmethod
    def install_dependencies(cls, *, check: bool = True) -> CompletedProcess[bytes]:
        """Install the dependencies."""
        from pyrig.src.project.mgt import PROJECT_MGT  # noqa: PLC0415

        return run_subprocess([PROJECT_MGT, "sync"], check=check)



"""Build utilities for creating and managing project builds.

This module provides functions for building and managing project artifacts,
including creating build scripts, configuring build environments, and
handling build dependencies. These utilities help with the packaging and
distribution of project code.
"""

import os
import platform
import shutil
import tempfile
from abc import ABC, abstractmethod
from pathlib import Path
from types import ModuleType

from PIL import Image

import pyrig
from pyrig import main
from pyrig.dev.artifacts import builder, resources
from pyrig.dev.artifacts.resources.resource import (
    get_all_resources_pkgs_from_deps_depen_on_dep,
)
from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.src.modules.class_ import (
    get_all_nonabst_subcls_from_mod_in_all_deps_depen_on_dep,
)
from pyrig.src.modules.module import (
    to_path,
)
from pyrig.src.modules.package import get_src_package


class Builder(ABC):
    """Base class for build scripts.

    Subclass this class and implement the get_artifacts method to create
    a build script for your project. The build method will be called
    automatically when the class is initialized. At the end of the file add
    if __name__ == "__main__":
        YourBuildClass()
    """

    ARTIFACTS_DIR_NAME = "dist"

    @classmethod
    @abstractmethod
    def create_artifacts(cls, temp_artifacts_dir: Path) -> None:
        """Build the project.

        This method should create all artifacts in the given folder.

        Returns:
            None
        """

    @classmethod
    def __init__(cls) -> None:
        """Initialize the build script."""
        cls.build()

    @classmethod
    def get_artifacts_dir(cls) -> Path:
        """Get the artifacts directory."""
        return Path(cls.ARTIFACTS_DIR_NAME)

    @classmethod
    def build(cls) -> None:
        """Build the project.

        This method is called by the __init__ method.
        It takes all the files and renames them with -platform.system()
        and puts them in the artifacts folder.
        """
        with tempfile.TemporaryDirectory() as temp_build_dir:
            temp_dir_path = Path(temp_build_dir)
            temp_artifacts_dir = cls.get_temp_artifacts_path(temp_dir_path)
            cls.create_artifacts(temp_artifacts_dir)
            artifacts = cls.get_temp_artifacts(temp_artifacts_dir)
            cls.rename_artifacts(artifacts)

    @classmethod
    def rename_artifacts(cls, artifacts: list[Path]) -> None:
        """Rename the artifacts in a non temporary folder."""
        artifacts_dir = cls.get_artifacts_dir()
        artifacts_dir.mkdir(parents=True, exist_ok=True)
        for artifact in artifacts:
            # rename the files with -platform.system()
            new_name = f"{artifact.stem}-{platform.system()}{artifact.suffix}"
            new_path = artifacts_dir / new_name
            shutil.move(str(artifact), str(new_path))

    @classmethod
    def get_temp_artifacts(cls, temp_artifacts_dir: Path) -> list[Path]:
        """Get the built artifacts."""
        paths = list(temp_artifacts_dir.glob("*"))
        if not paths:
            msg = f"Expected {temp_artifacts_dir} to contain files"
            raise FileNotFoundError(msg)
        return paths

    @classmethod
    def get_artifacts(cls) -> list[Path]:
        """Get the built artifacts."""
        return list(cls.get_artifacts_dir().glob("*"))

    @classmethod
    def get_temp_artifacts_path(cls, temp_dir: Path) -> Path:
        """Get the built artifacts."""
        path = temp_dir / cls.ARTIFACTS_DIR_NAME
        path.mkdir(parents=True, exist_ok=True)
        return path

    @classmethod
    def get_non_abstract_subclasses(cls) -> list[type["Builder"]]:
        """Get all non-abstract subclasses of Builder."""
        return get_all_nonabst_subcls_from_mod_in_all_deps_depen_on_dep(
            cls,
            pyrig,
            builder,
            discard_parents=True,
        )

    @classmethod
    def init_all_non_abstract_subclasses(cls) -> None:
        """Build all artifacts."""
        for builder_cls in cls.get_non_abstract_subclasses():
            builder_cls()

    @classmethod
    def get_app_name(cls) -> str:
        """Get the app name."""
        return PyprojectConfigFile.get_project_name()

    @classmethod
    def get_root_path(cls) -> Path:
        """Get the root path."""
        src_pkg = get_src_package()
        return to_path(src_pkg, is_package=True).resolve().parent

    @classmethod
    def get_main_path(cls) -> Path:
        """Get the main path."""
        return cls.get_src_pkg_path() / cls.get_main_path_from_src_pkg()

    @classmethod
    def get_resources_path(cls) -> Path:
        """Get the resources path."""
        return cls.get_src_pkg_path() / cls.get_resources_path_from_src_pkg()

    @classmethod
    def get_src_pkg_path(cls) -> Path:
        """Get the src package path."""
        return cls.get_root_path() / PyprojectConfigFile.get_package_name()

    @classmethod
    def get_main_path_from_src_pkg(cls) -> Path:
        """Get the main path.

        The path to main from the src package.
        """
        return to_path(main, is_package=False).relative_to(
            to_path(pyrig, is_package=True)
        )

    @classmethod
    def get_resources_path_from_src_pkg(cls) -> Path:
        """Get the resources path."""
        return to_path(resources, is_package=True).relative_to(
            to_path(pyrig, is_package=True)
        )


class PyInstallerBuilder(Builder):
    """Build the project with pyinstaller.

    Expects main.py in the src package.
    """

    @classmethod
    def create_artifacts(cls, temp_artifacts_dir: Path) -> None:
        """Build the project with pyinstaller."""
        from PyInstaller.__main__ import run  # noqa: PLC0415

        options = cls.get_pyinstaller_options(temp_artifacts_dir)
        run(options)

    @classmethod
    @abstractmethod
    def get_additional_resource_pkgs(cls) -> list[ModuleType]:
        """Get the resource packages.

        Override this methdod and return packages that conatin your resources.
        Those will be traversed and all files included in the build.
        The generated src_pkg/dev/artifacts/resources package is added by default.
        All resources folders from all pkgs depending on pyrig are also added by
        default.
        """

    @classmethod
    def get_default_additional_resource_pkgs(cls) -> list[ModuleType]:
        """Get the default additional resource packages."""
        return get_all_resources_pkgs_from_deps_depen_on_dep(dep=pyrig)

    @classmethod
    def get_all_resource_pkgs(cls) -> list[ModuleType]:
        """Get all resource packages."""
        return [
            *cls.get_default_additional_resource_pkgs(),
            *cls.get_additional_resource_pkgs(),
        ]

    @classmethod
    def get_add_datas(cls) -> list[tuple[Path, Path]]:
        """Get the add data paths.

        Returns:
            list[tuple[Path, Path]]: List of tuples with the source path
                and the destination path.
        """
        add_datas: list[tuple[Path, Path]] = []
        resources_pkgs = cls.get_all_resource_pkgs()
        for pkg in resources_pkgs:
            pkg_path = Path(pkg.__path__[0])
            # get the root of the pkg, which will be
            # the path remove suufix get_resources_path_from_src_pkg
            pkg_root = Path(
                pkg_path.as_posix().removesuffix(
                    cls.get_resources_path_from_src_pkg().as_posix()
                )
            ).parent
            for path in [pkg_path, *pkg_path.rglob("*")]:
                if path.is_file():
                    continue
                dest = path.relative_to(pkg_root)
                add_datas.append((path, dest))
        return add_datas

    @classmethod
    def get_pyinstaller_options(cls, temp_artifacts_dir: Path) -> list[str]:
        """Get the pyinstaller options."""
        temp_dir = temp_artifacts_dir.parent

        options = [
            str(cls.get_main_path()),
            "--name",
            cls.get_app_name(),
            "--clean",
            "--noconfirm",
            "--onefile",
            "--noconsole",
            "--workpath",
            str(cls.get_temp_workpath(temp_dir)),
            "--specpath",
            str(cls.get_temp_specpath(temp_dir)),
            "--distpath",
            str(cls.get_temp_distpath(temp_dir)),
            "--icon",
            str(cls.get_app_icon_path(temp_dir)),
        ]
        for src, dest in cls.get_add_datas():
            options.extend(["--add-data", f"{src}{os.pathsep}{dest}"])
        return options

    @classmethod
    def get_temp_distpath(cls, temp_dir: Path) -> Path:
        """Get the distpath option."""
        return cls.get_temp_artifacts_path(temp_dir)

    @classmethod
    def get_temp_workpath(cls, temp_dir: Path) -> Path:
        """Get the workpath option."""
        path = temp_dir / "workpath"
        path.mkdir(parents=True, exist_ok=True)
        return path

    @classmethod
    def get_temp_specpath(cls, temp_dir: Path) -> Path:
        """Get the specpath option."""
        path = temp_dir / "specpath"
        path.mkdir(parents=True, exist_ok=True)
        return path

    @classmethod
    def get_app_icon_path(cls, temp_dir: Path) -> Path:
        """Get the app icon path."""
        if platform.system() == "Windows":
            return cls.convert_png_to_format("ico", temp_dir)
        if platform.system() == "Darwin":
            return cls.convert_png_to_format("icns", temp_dir)
        return cls.convert_png_to_format("png", temp_dir)

    @classmethod
    def convert_png_to_format(cls, file_format: str, temp_dir_path: Path) -> Path:
        """Convert a png to a format."""
        output_path = temp_dir_path / f"icon.{file_format}"
        png_path = cls.get_app_icon_png_path()
        img = Image.open(png_path)
        img.save(output_path, format=file_format.upper())
        return output_path

    @classmethod
    def get_app_icon_png_path(cls) -> Path:
        """Get the app icon path.

        Default is under dev/artifacts/resources folder as icon.png
        You can override this method to change the icon location.
        """
        return cls.get_resources_path() / "icon.png"
